#!/usr/bin/env python
import os
import sys
import argparse
import yaml
from relecov_tools.download_manager import DownloadManager
from relecov_tools.dataprocess_wrapper import ProcessWrapper


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--download_option",
        type=str,
        help="Download option",
    )
    parser.add_argument("-t", "--target_folders", type=str, help="Target folders")
    args = parser.parse_args()

    val_dict = {
        "user": os.environ["TEST_USER"],
        "password": os.environ["TEST_PASSWORD"],
        "download_option": args.download_option,
        "output_dir": os.environ["OUTPUT_LOCATION"],
        "target_folders": args.target_folders,
        "sftp_port": os.environ["TEST_PORT"],
    }
    prepare_remote_test(**val_dict)


def generate_config_yaml(download_option, target_folders):
    """Generate the wrapper_config.yaml file with the desired structure."""
    config_data = {
        "download": {
            "user": "",
            "password": "",
            "download_option": download_option,
            "target_folders": target_folders,
            "subfolder": "RELECOV",
        },
        "read-lab-metadata": {
            "metadata_file": "tests/data/read_lab_metadata/metadata_lab_test.xlsx",
            "sample_list_file": "tests/data/read_lab_metadata/samples_data_test.json",
        },
        "validate": {
            "json_schema_file": "relecov_tools/schema/relecov_schema.json",
            "registry": "tests/data/map_validate/unique_sampleid_registry.json",
        },
    }

    with open("wrapper_config.yaml", "w") as file:
        yaml.dump(config_data, file, default_flow_style=False)

    return "wrapper_config.yaml"


def prepare_remote_test(**kwargs):
    # First clean the repository.
    print("Initating sftp module")

    download_manager = DownloadManager(
        user=kwargs["user"],
        password=kwargs["password"],
        conf_file=None,
        download_option=kwargs["download_option"],
        output_dir=kwargs["output_dir"],
        target_folders=kwargs["target_folders"],
        subfolder="RELECOV",
    )
    print("Openning connection to sftp")
    download_manager.relecov_sftp.sftp_port = os.environ["TEST_PORT"]
    if not download_manager.relecov_sftp.open_connection():
        print("Could not open connection to remote sftp")
        sys.exit(1)

    print("Cleaning folders inside RELECOV")
    base_paths = ["COD-test-1/RELECOV", "COD-test-2/RELECOV"]
    for base_path in base_paths:
        remote_folders = download_manager.relecov_sftp.list_remote_folders(
            base_path, recursive=True
        )
        remote_folders = sorted(remote_folders, reverse=True)
        for folder in remote_folders:
            folder = folder.replace("./", "")
            if folder == base_path:
                continue

            filelist = download_manager.relecov_sftp.get_file_list(folder)
            for file in filelist:
                print(f"Removing file {file} in {folder}")
                download_manager.relecov_sftp.remove_file(file)

            if "tmp_processing" in folder or "invalid_samples" in folder:
                try:
                    print(f"Removing special folder {folder}")
                    download_manager.relecov_sftp.remove_dir(folder)
                except Exception as e:
                    print(f"Could not remove folder {folder}: {e}")

        filelist_base = download_manager.relecov_sftp.get_file_list(base_path)
        for file in filelist_base:
            print(f"Removing file {file} in {base_path}")
            download_manager.relecov_sftp.remove_file(file)

    data_loc = "tests/data/sftp_handle"
    folder_files_dict = {folder: files for folder, _, files in os.walk(data_loc)}
    print("Uploading files to sftp...")
    for folder, files in folder_files_dict.items():
        if "datatest" in folder:
            remote_dir = "COD-test-1/RELECOV"
        elif "empty_test" in folder:
            remote_dir = "COD-test-2/RELECOV"
        else:
            continue
        print(f"Uploading files from {folder}")
        for file in files:
            remote_path = os.path.join(remote_dir, file)
            local_path = os.path.join(os.path.abspath(folder), file)
            download_manager.relecov_sftp.upload_file(local_path, remote_path)

    download_manager.relecov_sftp.close_connection()

    print("Create Intermediate YML")
    conf_file = generate_config_yaml(  # Generate the YAML config file
        kwargs["download_option"],
        kwargs["target_folders"],
    )

    print("Initiating ProcessWrapper")
    wrapper_manager = ProcessWrapper(  # Initialize ProcessWrapper with the config file
        config_file=conf_file,
        output_dir=kwargs["output_dir"],
    )

    print("Update Wrapper params")
    wrapper_manager.download_params = {  # Set download parameters
        "user": os.environ["TEST_USER"],
        "password": os.environ["TEST_PASSWORD"],
        "download_option": kwargs["download_option"],
        "output_dir": kwargs["output_dir"],
        "target_folders": kwargs["target_folders"],
        "sftp_port": kwargs["sftp_port"],
        "subfolder": "RELECOV",
    }

    def test_wrapper(wrapper_manager):
        wrapper_manager.run_wrapper()  # Execute the wrapper process

    test_wrapper(wrapper_manager)  # Run the wrapper function


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import re
import yaml
import os
import inspect
import rich.console
from collections import defaultdict
from relecov_tools.download_manager import DownloadManager
from relecov_tools.read_lab_metadata import RelecovMetadata
from relecov_tools.json_validation import SchemaValidation
from relecov_tools.base_module import BaseModule
import relecov_tools.utils

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class ProcessWrapper(BaseModule):
    """
    Always fill all the arguments for the class in the config file, leave its value
    if you dont want to use that argument e.g.(target_folders:  ) -> (target_folders = None)
    """

    def __init__(self, config_file: str = None, output_dir: str = None):
        super().__init__(output_dir=output_dir, called_module="wrapper")
        if not os.path.isdir(str(output_dir)):
            raise FileNotFoundError(f"Output folder {output_dir} is not valid")
        else:
            self.output_dir = output_dir
        if not os.path.isfile(str(config_file)):
            raise FileNotFoundError(f"Config file {config_file} is not a file")
        else:
            try:
                self.config_data = relecov_tools.utils.read_yml_file(config_file)
                # Config file should include a key
            except yaml.YAMLError as e:
                raise yaml.YAMLError(f"Invalid config file: {e}")
        output_regex = ("out_folder", "output_dir", "output_location")
        for key, val in self.config_data.items():
            for arg in output_regex:
                if val == arg:
                    self.config_data[key] = self.output_dir
        self.wrapper_logsum = self.parent_log_summary(
            output_dir=os.path.join(self.output_dir)
        )
        self.config_data["download"].update({"output_dir": output_dir})
        if "subfolder" not in self.config_data["download"]:
            self.config_data["download"].update(
                {"subfolder": "RELECOV"}
            )  # If subfolder is not defined or None, it is set automatically as RELECOV
            self.log.warning("Subfolder not provided. Set to as RELECOV by default")
            stderr.print("[yellow]Subfolder not provided. Set to RELECOV by default")
        self.download_params = self.clean_module_params(
            "DownloadManager", self.config_data["download"]
        )
        self.readmeta_params = self.clean_module_params(
            "RelecovMetadata", self.config_data["read-lab-metadata"]
        )
        self.validate_params = self.clean_module_params(
            "SchemaValidation", self.config_data["validate"]
        )

    def clean_module_params(self, module, params):
        active_module = eval(module)
        module_args = inspect.getfullargspec(active_module.__init__)[0]
        module_args.remove("self")
        module_valid_params = {x: y for x, y in params.items() if x in module_args}
        if not module_valid_params:
            self.log.error(f"Invalid params for {module} in config file")
            stderr.print(f"[red]Invalid params for {module} in config file")
            raise ValueError(f"Invalid params for {module}. Use: {module_args}")
        return module_valid_params

    def exec_download(self, download_params):
        if "sftp_port" in download_params:
            sftp_port = download_params.pop("sftp_port", None)
        else:
            sftp_port = None
        download_manager = DownloadManager(**download_params)
        if sftp_port is not None:
            download_manager.relecov_sftp.sftp_port = int(sftp_port)
            print(f"SFTP port assigned: {download_manager.relecov_sftp.sftp_port}")
        download_manager.defer_cleanup = True
        download_manager.execute_process()
        finished_folders = download_manager.finished_folders
        download_logs = self.wrapper_logsum.prepare_final_logs(
            logs=download_manager.logsum.logs
        )
        self.download_manager = download_manager
        return finished_folders, download_logs

    def exec_read_metadata(self, readmeta_params):
        read_metadata = RelecovMetadata(**readmeta_params)
        read_metadata.create_metadata_json()
        read_meta_logs = self.wrapper_logsum.prepare_final_logs(
            logs=read_metadata.logsum.logs
        )
        return read_meta_logs

    def exec_validation(self, validate_params):
        validate_proccess = SchemaValidation(**validate_params)
        valid_json_data, invalid_json = validate_proccess.validate()
        validate_logs = self.wrapper_logsum.prepare_final_logs(
            logs=validate_proccess.logsum.logs
        )
        return valid_json_data, invalid_json, validate_logs

    def process_folder(self, finished_folders, key, folder_logs):
        """Executes read-lab-metadata and validation process for the given downloaded folder.
        Merges all the log summaries generated with the ones from download process, creates
        an excel file with custom format and uploads it back to its remote sftp folder.
        Also uploads the files that failed validation back to the remote sftp folder.
        Finally. It cleans all the remote remaining files if the process was successful.

        Args:
            finished_folders (dict(str:list)): Dictionary which includes the names
            of the remote folders processed during download and the successfull files for each
            key (str): Name of the folder to process in remote sftp, same name as the one
            included in the log_summary from download process.
            folder_logs (dict): Download log_summary corresponding to the processed folder

        Raises:
            ValueError: If folder_logs dont include a path to the local folder
            ValueError: If folder is not found in remote sftp or more than 1
            ValueError: If no samples/files are found for the folder after download
            ValueError: If no metadata json file is found after read-lab-metadata

        Returns:
            merged_logs (dict): Dictionary which includes the logs from all processes
        """

        def upload_files_from_json(invalid_json, remote_dir):
            """Upload the files in a given json with samples metadata"""
            for sample in invalid_json:
                local_dir = sample.get("sequence_file_path_R1")
                # files_keys = [key for key in sample.keys() if "_file_" in key]
                sample_files = (
                    sample.get("sequence_file_R1"),
                    sample.get("sequence_file_R2"),
                )
                ftp_files = self.download_manager.relecov_sftp.get_file_list(remote_dir)
                uploaded_files = []
                for file in sample_files:
                    if not file or file in ftp_files:
                        continue
                    loc_path = os.path.join(local_dir, file)
                    sftp_path = os.path.join(remote_dir, file)
                    self.log.info("Uploading %s to remote %s" % (loc_path, remote_dir))
                    uploaded = self.download_manager.relecov_sftp.upload_file(
                        loc_path, sftp_path
                    )
                    if not uploaded:
                        err = f"Could not upload {loc_path} to {remote_dir}"
                        self.wrapper_logsum.add_error(sample=sample, entry=err)
                    else:
                        uploaded_files.append(file)
            return uploaded_files

        local_folder = folder_logs.get("path")
        self.log.info(f"Working in downloaded local folder {local_folder}")
        if not local_folder:
            raise ValueError(f"Couldnt find local path for {key} in log after download")
        files = [os.path.join(local_folder, file) for file in os.listdir(local_folder)]
        try:
            metadata_file = [x for x in files if re.search("lab_metadata.*.xlsx", x)][0]
            samples_file = [x for x in files if re.search("samples_data.*.json", x)][0]
        except IndexError:
            raise ValueError("No metadata/samples files found after download")
        self.readmeta_params.update(
            {
                "metadata_file": metadata_file,
                "sample_list_file": samples_file,
                "output_dir": local_folder,
            }
        )
        read_meta_logs = self.exec_read_metadata(self.readmeta_params)
        metadata_json = [
            x
            for x in os.listdir(local_folder)
            if re.search(r"lab_metadata.*\.json$", x)
            and not x.endswith("_log_summary.json")
        ]
        if not metadata_json:
            raise ValueError("No metadata json found after read-lab-metadata")
        self.validate_params.update(
            {
                "json_file": os.path.join(local_folder, metadata_json[0]),
                "metadata": metadata_file,
                "output_dir": local_folder,
            }
        )
        valid_json_data, invalid_json, validate_logs = self.exec_validation(
            self.validate_params
        )
        merged_logs = self.wrapper_logsum.merge_logs(
            key_name=key, logs_list=[{key: folder_logs}, read_meta_logs, validate_logs]
        )
        stderr.print(f"[green]Merged logs from all processes in {local_folder}")
        self.log.info(f"Merged logs from all processes in {local_folder}")
        subfolder = getattr(self.download_manager, "subfolder", None)
        if subfolder and subfolder not in key:
            main_folder = os.path.join(key, subfolder)
        else:
            main_folder = key
        sftp_dirs = self.download_manager.relecov_sftp.list_remote_folders(main_folder)
        sftp_dirs_paths = [os.path.join(main_folder, d) for d in sftp_dirs]
        valid_dirs = [d for d in sftp_dirs_paths if d in finished_folders.keys()]

        # As all folders are merged into one during download, there should only be 1 folder
        if (not valid_dirs or len(valid_dirs) >= 2) and invalid_json:
            # If all samples were valid during download and download_clean is used, the original folder might have been deleted
            self.log.warning(
                "Couldnt find %s folder in remote sftp. Creating new one", main_folder
            )
            remote_dir = os.path.join(main_folder, self.batch_id + "_invalid_samples")
            self.download_manager.relecov_sftp.make_dir(remote_dir)
        else:
            remote_dir = valid_dirs[0]
            stderr.print(
                f"[blue]Cleaning successfully validated files from {remote_dir}"
            )
            self.log.info(
                f"Cleaning successfully validated files from remote dir: {remote_dir}"
            )
            file_fields = ("sequence_file_R1", "sequence_file_R2")
            valid_sampfiles = [f.get(v) for v in file_fields for f in valid_json_data]
            remote_files = self.download_manager.relecov_sftp.get_file_list(remote_dir)
            valid_files = [f for f in remote_files if f in valid_sampfiles]
            self.download_manager.delete_remote_files(remote_dir, files=valid_files)
            self.download_manager.delete_remote_files(remote_dir, skip_seqs=True)
            self.download_manager.clean_remote_folder(remote_dir)
        if invalid_json:
            logtxt = f"Found {len(invalid_json)} invalid samples in {key}"
            self.wrapper_logsum.add_warning(key=key, entry=logtxt)
            if "tmp_processing" in remote_dir:
                renamed_dir = remote_dir.replace("tmp_processing", "invalid_samples")
                remote_dir = renamed_dir

            invalid_metadata_files = [
                x
                for x in os.listdir(local_folder)
                if re.search(r"invalid_lab_metadata.*\.xlsx", x)
            ]

            if not invalid_metadata_files:
                self.log.warning("No invalid_lab_metadata_*.xlsx file found to upload")
            else:
                if len(invalid_metadata_files) > 1:
                    self.log.warning(
                        "Multiple invalid_lab_metadata_*.xlsx files found: %s. Uploading the first one.",
                        invalid_metadata_files,
                    )
                invalid_metadata_path = os.path.join(
                    local_folder, invalid_metadata_files[0]
                )
                sftp_path = os.path.join(
                    remote_dir, os.path.basename(invalid_metadata_path)
                )
                self.log.info("Uploading invalid files and metadata to %s", remote_dir)
                stderr.print(
                    f"[blue]Uploading invalid files and metadata to {remote_dir}"
                )
                self.download_manager.relecov_sftp.upload_file(
                    invalid_metadata_path, sftp_path
                )
            # Upload all the files that failed validation process back to sftp
            upload_files_from_json(invalid_json, remote_dir)
        else:
            self.log.info("No invalid samples in %s", key)
            stderr.print(f"[green]No invalid samples were found for {key} !!!")
        log_filepath = os.path.join(local_folder, str(key) + "_metadata_report.json")
        self.wrapper_logsum.create_error_summary(
            called_module="metadata",
            filepath=log_filepath,
            logs=merged_logs,
            to_excel=True,
        )
        xlsx_report_files = [
            f for f in os.listdir(local_folder) if re.search("metadata_report.xlsx", f)
        ]
        if xlsx_report_files and invalid_json:
            self.log.info("Uploading %s xlsx report to remote %s" % (key, remote_dir))
            local_xlsx = os.path.join(local_folder, xlsx_report_files[0])
            remote_xlsx = os.path.join(remote_dir, xlsx_report_files[0])
            up = self.download_manager.relecov_sftp.upload_file(local_xlsx, remote_xlsx)
            if not up:
                self.log.error(
                    "Could not upload %s report to remote %s" % (key, local_folder)
                )
        elif not xlsx_report_files and invalid_json:
            self.log.error(
                "Could not find xlsx report for %s in %s" % (key, local_folder)
            )
        return merged_logs

    def run_wrapper(self):
        """Execute each given process in config file sequentially, starting with download.
        Once the download has finished, each downloaded folder is processed using read-lab-metadata
        and validation modules. The logs from each module are merged into a single log-summary.
        These merged logs are then used to create an excel report of all the processes
        """
        self.log.info("Starting with wrapper")
        finished_folders, download_logs = self.exec_download(self.download_params)
        self.set_batch_id(self.download_manager.batch_id)
        if not finished_folders:
            stderr.print("[red]No valid folders found to process")
            self.log.error("No valid folders found to process")
            raise FileNotFoundError("No valid folders found to process")
        stderr.print(f"[blue]Processing {len(finished_folders)} downloaded folders...")
        self.log.info(f"Processing {len(finished_folders)} downloaded folders.")
        counter = 0
        for key, folder_logs in download_logs.items():
            folder = folder_logs.get("path")
            if not folder:
                self.log.error(f"Skipped folder {key}. Logs do not include path field")
                continue
            if not folder_logs.get("valid"):
                self.log.error(f"Folder {key} is set as invalid in logs. Skipped.")
                continue
            counter += 1
            logtxt = "Processing folder %s/%s: %s with local path %s"
            self.log.info(logtxt % (counter, str(len(finished_folders)), key, folder))
            stderr.print(logtxt % (counter, str(len(finished_folders)), key, folder))
            try:
                merged_logs = self.process_folder(finished_folders, key, folder_logs)
            except (FileNotFoundError, ValueError) as e:
                self.log.error(f"Could not process folder {key}: {e}")
                folder_logs["errors"].append(f"Could not process folder {key}: {e}")
                log_filepath = os.path.join(
                    folder, self.tag_filename(f"wrapper_{key}.json")
                )
                log_filepath = log_filepath.replace(".json", "_log_summary.json")
                self.parent_create_error_summary(
                    called_module="metadata",
                    filepath=log_filepath,
                    logs={key: folder_logs},
                    to_excel=False,
                )
                continue
            self.wrapper_logsum.logs[key] = merged_logs[key]

        self.parent_create_error_summary(
            called_module="wrapper",
            to_excel=True,
        )

        # Logging wrapper stats
        labs = {folder.split("/")[0] for folder in finished_folders}
        num_labs = len(labs)
        samples_per_lab = defaultdict(int)
        for folder, files in finished_folders.items():
            lab = folder.split("/")[0]
            seq_files = [f for f in files if not f.endswith(".xlsx")]
            samples_per_lab[lab] += len(seq_files)
        total_count = sum(samples_per_lab.values())

        stderr.print("[blue] --------------------")
        stderr.print("[blue] WRAPPER SUMMARY")
        stderr.print("[blue] --------------------")
        self.log.info(f"Number of processed laboratories: {num_labs}")
        stderr.print(f"[blue]Number of processed laboratories: {num_labs}")
        self.log.info(f"Total number of processed files: {total_count}")
        stderr.print(f"[blue]Total number of processed files: {total_count}")
        stderr.print("[blue]Processed files for each laboratory")
        for lab, count in samples_per_lab.items():
            self.log.info(f"    {lab}: {count} files")
            stderr.print(f"[blue]    {lab}: {count} files")

        return



# =============================================================
# INTRODUCTION

# This script is run in order to organise data from the RELECOV analyses according to their corresponding epidemiological weeks.
# By default, all data will be stored in a folder called surveillance_files.
# Within this folder, different subfolders will be created, each one referring to a certain epidemiological week.
# Inside each subfolder, the following items are stored:
# - epidemiological_data.xlsx: an excel file containing lineage information for all the samples from a given week. This information is also aggregated in another sheet.
# - variant_data.csv: a .csv file containing information regarding the variants identified for all the samples associated to a given week.
# - consensus_files: a subfolder containing all the consensus.fa files obtained after the analysis of samples.

# =============================================================

import os
import json
import argparse
import shutil
import pandas as pd
from datetime import datetime


# Function to determine the epidemiological week associated to a certain date.
def get_epi_week(date_str):
    date = datetime.strptime(date_str, "%Y-%m-%d")
    year, week, weekday = date.isocalendar()
    return f"{year}-{week:02d}"


# Function to search .json files in the paths indicated in the provided .txt files (specifically, bioinfo_lab_metadata and long_table .json files), read them, extract the relevant information and generate tables.
def process_json_files(
    input_dir=None,
    metadata_list=None,
    long_table_list=None,
    metadata_file=None,
    long_table_file=None,
    output_dir="surveillance_files",
    specified_week=None,
    copy_fasta=False,
):
    os.makedirs(output_dir, exist_ok=True)

    bioinfo_files = []
    if metadata_list:
        with open(metadata_list, "r", encoding="utf-8") as f:
            bioinfo_files = [line.strip() for line in f if line.strip()]
    elif metadata_file:
        bioinfo_files = [metadata_file]
    elif input_dir:
        bioinfo_files = [
            os.path.join(input_dir, filename)
            for filename in os.listdir(input_dir)
            if filename.startswith("bioinfo_lab_metadata_")
            and filename.endswith(".json")
        ]

    long_table_files = []
    if long_table_list:
        with open(long_table_list, "r", encoding="utf-8") as f:
            long_table_files = [line.strip() for line in f if line.strip()]
    elif long_table_file:
        long_table_files = [long_table_file]
    elif input_dir:
        long_table_files = [
            os.path.join(input_dir, filename)
            for filename in os.listdir(input_dir)
            if filename.startswith("long_table_") and filename.endswith(".json")
        ]

    all_data = []
    fa_files = []
    sample_variant_data = {}

    # Processing of bioinfo_lab_metadata_*.json.
    for filepath in bioinfo_files:
        if not os.path.exists(filepath):
            print(
                f"Warning! The file {filepath} could not be found. Please make sure the path is correct."
            )
            continue

        with open(filepath, "r", encoding="utf-8") as file:
            try:
                data = json.load(file)
                for sample in data:
                    if "sample_collection_date" in sample:
                        week = get_epi_week(sample["sample_collection_date"])
                        if specified_week and week != specified_week:
                            continue

                        analysis_date = sample.get("bioinformatics_analysis_date", "-")

                        all_data.append(
                            {
                                "HOSPITAL_ID": sample.get(
                                    "submitting_institution_id", "-"
                                ),
                                "HOSPITAL": sample.get("collecting_institution", "-"),
                                "PROVINCE": sample.get("geo_loc_region", "-"),
                                "ANALYSIS_DATE": analysis_date,
                                "PANGOLIN_SOFTWARE_VERSION": sample.get(
                                    "lineage_assignment_software_version", "-"
                                ),
                                "PANGOLIN_DATABASE_VERSION": sample.get(
                                    "lineage_assignment_database_version", "-"
                                ),
                                "SAMPLE_ID": str(
                                    sample.get("sequencing_sample_id", "-")
                                ),  # str to prevent from having issues between excel and pandas with digital and characters
                                "SAMPLE_COLLECTION_DATE": sample.get(
                                    "sample_collection_date", "-"
                                ),
                                "LINEAGE": sample.get("lineage_assignment", "-"),
                                "WEEK": week,
                            }
                        )

                        # Search of consensus.fa files.
                        fa_path = sample.get("consensus_sequence_filepath")
                        if copy_fasta and fa_path and os.path.exists(fa_path):
                            fa_files.append((fa_path, week))

            except json.JSONDecodeError:
                print(
                    f"Error! Could not read {filepath} properly, please make sure the file is not corrupt."
                )

    if not all_data:
        print("No bioinfo_lab_metadata_*.json files were found.")
        return

    # Processing of long_table_*.json.
    for filepath in long_table_files:
        if not os.path.exists(filepath):
            print(
                f"Warning! The file {filepath} could not be found. Please check the path is correct."
            )
            continue

        with open(filepath, "r", encoding="utf-8") as file:
            try:
                data = json.load(file)
                for sample in data:
                    sample_id = sample.get("sample_name")
                    if sample_id:
                        sample_variant_data[sample_id] = sample
            except json.JSONDecodeError:
                print(
                    f"Error! Could not read {filepath} properly, please make sure the file is not corrupt."
                )

    if not sample_variant_data:
        print("No long_table_*.json files were found.")
        return

    df = pd.DataFrame(all_data)
    weeks = df["WEEK"].unique()

    # Creation of epidemiological-week folders, the .xlsx file with lineage information per week and the .csv file with the variant information per week.
    for week in weeks:
        week_dir = os.path.join(output_dir, week)
        os.makedirs(week_dir, exist_ok=True)

        week_df = df[df["WEEK"] == week]
        excel_file = os.path.join(week_dir, "epidemiological_data.xlsx")

        existing_sample_ids = set()
        existing_week_df = pd.DataFrame()

        if os.path.exists(excel_file):
            # Check if xlsx file exists, read data and list the SAMPLE_ID
            with pd.ExcelFile(excel_file) as reader:
                existing_week_df = reader.parse(
                    "per_sample_data", dtype=str
                )  # str ensures no mofications are done between what excel has and what the dataframe reads
                existing_sample_ids = set(existing_week_df["SAMPLE_ID"])

        # only new samples are added
        week_df = week_df[~week_df["SAMPLE_ID"].astype(str).isin(existing_sample_ids)]

        if week_df.empty:
            print(f"No new samples for week {week}. Skipping.")
            continue

        # concatenate new records to those already in the excel file
        week_df = pd.concat([existing_week_df, week_df], ignore_index=True)
        aggregated_df = (
            week_df.groupby("LINEAGE").size().reset_index(name="NUMBER_SAMPLES")
        )

        with pd.ExcelWriter(excel_file) as writer:
            week_df.to_excel(writer, sheet_name="per_sample_data", index=False)
            aggregated_df.to_excel(writer, sheet_name="aggregated_data", index=False)

        print(f"Tables were stored in {week_dir}")

        # Copy of the consensus.fa files into a subfolder called consensus_files.
        if copy_fasta:
            consensus_dir = os.path.join(week_dir, "consensus_files")
            os.makedirs(consensus_dir, exist_ok=True)
            for fa_path, week_fa in fa_files:
                if week_fa == week:
                    dest_path = os.path.join(consensus_dir, os.path.basename(fa_path))
                    shutil.copy(fa_path, dest_path)
            print("Copy of consensus.fa files completed successfully")

        # Generation of the .csv files with variant data.
        variant_data = []
        for _, row in week_df.iterrows():
            sample_id = row["SAMPLE_ID"]
            if sample_id in sample_variant_data:
                variant_entries = sample_variant_data[sample_id].get("variants", [])
                for variant in variant_entries:
                    variant_data.append(
                        {
                            "SAMPLE": variant.get("sample", "-"),
                            "CHROM": variant.get("chromosome", "-"),
                            "POS": variant.get("pos", "-"),
                            "ALT": variant.get("alt", "-"),
                            "REF": variant.get("ref", "-"),
                            "FILTER": variant.get("Filter", "-"),
                            "DP": variant.get("dp", "-"),
                            "REF_DP": variant.get("ref_dp", "-"),
                            "ALT_DP": variant.get("alt_dp", "-"),
                            "AF": variant.get("af", "-"),
                            "GENE": variant.get("gene", "-"),
                            "EFFECT": variant.get("effect", "-"),
                            "HGVS_C": variant.get("hgvs_c", "-"),
                            "HGVS_P": variant.get("hgvs_p", "-"),
                            "HGVS_P_1LETTER": variant.get("hgvs_p_1_letter", "-"),
                            "CALLER": variant.get("caller", "-"),
                            "LINEAGE": variant.get("lineage", "-"),
                        }
                    )

        if variant_data:
            variant_df = pd.DataFrame(variant_data)
            variant_csv = os.path.join(week_dir, "variant_data.csv")

        if os.path.exists(variant_csv):
            existing_df = pd.read_csv(variant_csv, dtype=str)
            final_df = pd.concat([existing_df, variant_df], ignore_index=True)
            final_variant_df = final_df.drop_duplicates()

        else:
            final_variant_df = variant_df

        final_variant_df.to_csv(variant_csv, index=False)
        print(f"Variant data stored in {variant_csv}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="JSON files are processed in order to asssign epidemiological weeks to each  analysed sample and to generate tables with lineage and variant information for each sample "
    )
    parser.add_argument(
        "-i",
        "--input",
        help=" Directoy containing files called bioinfo_lab_metadata_*.json and long_table_*.json (both JSON files must be in the same directory)",
    )
    parser.add_argument(
        "-b",
        "--metadata-list",
        help="File .txt with paths to those JSON files to be processed for generating the table with information on aggregated and non-aggregated data (bioinfo_lab_metadata_*.json)",
    )
    parser.add_argument(
        "-l",
        "--long-table-list",
        help="File .txt with paths to those JSON files to be processed for generating the variant report in csv format (long_table_*.json)",
    )
    parser.add_argument(
        "-m",
        "--metadata-file",
        help="Direct path to a single bioinfo_lab_metadata_*.json file to process and to add to the tables if they already exist",
    )
    parser.add_argument(
        "-t",
        "--long-table-file",
        help="Direct path to a single long_table_*.json file to process and to add to the tables if they already exist",
    )
    parser.add_argument(
        "-o",
        "--output",
        default="surveillance_files",
        help="Directory where the generated tables are saved (surveillance_files, by default)",
    )
    parser.add_argument(
        "-w",
        "--week",
        help="Epidemiological week of intestest (format: YYYY-WW)",
    )
    parser.add_argument(
        "-c",
        "--copy-fasta",
        action="store_true",
        help="Copy files consensus.fa in a subdirectory called consensus_files (this argument must be provided to enable the feature)",
    )

    args = parser.parse_args()

    # Validate that either input directory or file lists or individual files are provided
    if not (args.input or args.metadata_list or args.metadata_file):
        parser.error(
            "Either --input, --metadata-list, or --metadata-file must be provided"
        )
    if not (args.input or args.long_table_list or args.long_table_file):
        parser.error(
            "Either --input, --long-table-list, or --long-table-file must be provided"
        )

    process_json_files(
        args.input,
        args.metadata_list,
        args.long_table_list,
        args.metadata_file,
        args.long_table_file,
        args.output,
        args.week,
        args.copy_fasta,
    )



#!/usr/bin/env python
import sys
import os
import re
import json
import rich.console
import time

import relecov_tools.utils
from relecov_tools.config_json import ConfigJson
from relecov_tools.rest_api import RestApi
from relecov_tools.base_module import BaseModule

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class UpdateDatabase(BaseModule):
    def __init__(
        self,
        user=None,
        password=None,
        json=None,
        type=None,
        platform=None,
        server_url=None,
        full_update=False,
        long_table=None,
    ):
        if json is None:
            json = relecov_tools.utils.prompt_path(
                msg="Select the json file which have the data to map"
            )
        json_dir = os.path.dirname(os.path.realpath(json))
        super().__init__(output_dir=json_dir, called_module="update-db")
        # Get the user and password for the database
        if user is None:
            user = relecov_tools.utils.prompt_text(
                msg="Enter username for upload data to server"
            )
        self.user = user
        if password is None:
            password = relecov_tools.utils.prompt_text(msg="Enter credential password")
        self.passwd = password
        # get the default coonfiguration used the instance
        self.config_json = ConfigJson()

        if not os.path.isfile(json):
            self.log.error("json data file %s does not exist ", json)
            stderr.print(f"[red] json data file {json} does not exist")
            sys.exit(1)
        self.json_data = relecov_tools.utils.read_json_file(json)
        batch_id = self.get_batch_id_from_data(self.json_data)
        self.set_batch_id(batch_id)
        for row in self.json_data:
            for key, value in row.items():
                if not isinstance(value, str):
                    row[key] = str(value)
        self.json_file = json
        schema = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "schema",
            self.config_json.get_topic_data("json_schemas", "relecov_schema"),
        )
        self.schema = relecov_tools.utils.read_json_file(schema)
        if full_update is True:
            if not os.path.isfile(long_table):
                raise ValueError(
                    f"Provided long_table file does not exist {long_table}"
                )
            self.full_update = True
            self.server_url = None
            self.long_table_file = os.path.realpath(long_table)
        else:
            self.full_update = False
            if type is None:
                type = relecov_tools.utils.prompt_selection(
                    "Select:",
                    ["sample", "bioinfodata", "variantdata"],
                )
            self.type_of_info = type
            if self.type_of_info == "variantdata":
                if long_table is None:
                    json = relecov_tools.utils.prompt_path(
                        msg="Select the json file which have the data to map"
                    )
                    if not os.path.isfile(long_table):
                        raise ValueError(
                            f"Provided long_table file does not exist {long_table}"
                        )
                self.long_table_file = os.path.realpath(long_table)
            # collect data for plarform to upload data
            if platform is None:
                platform = relecov_tools.utils.prompt_selection(
                    "Select:",
                    ["iskylims", "relecov"],
                )
            self.platform = platform
            if server_url is None:
                self.server_url = server_url
        # Get configuration settings for upload database
        try:
            self.platform_settings = self.config_json.get_topic_data(
                "upload_database", "platform"
            )
        except KeyError as e:
            logtxt = f"Unable to fetch parameters for {platform} {e}"
            stderr.print(f"[red]{logtxt}")
            self.log.error(logtxt)
            sys.exit(1)
        # create the instance for logging the summary information
        lab_code = json_dir.split("/")[-2]
        self.logsum = self.parent_log_summary(
            output_dir=json_dir, lab_code=lab_code, path=json_dir
        )

    def get_schema_ontology_values(self):
        """Read the schema and extract the values of ontology with the label"""
        ontology_dict = {}
        for prop, values in self.schema["properties"].items():
            if "ontology" in values and values["ontology"] != "":
                ontology_dict[values["ontology"]] = prop
        return ontology_dict

    def map_iskylims_sample_fields_values(self, sample_fields, s_project_fields):
        """Map the values to the properties send to databasee
        in json schema based on label
        """
        sample_list = []
        s_fields = list(sample_fields.keys())
        for row in self.json_data:
            s_dict = {}
            for sfield in s_fields:
                if sfield not in row.keys():
                    s_dict[sample_fields[sfield]] = "Not Provided"
                else:
                    value = row[sfield]
                    if isinstance(value, str):
                        found_ontology = re.search(r"(.+) \[\w+:.*", value)
                        if found_ontology:
                            # remove the ontology data from item value
                            value = found_ontology.group(1)
                    s_dict[sample_fields[sfield]] = value
            for pfield in s_project_fields:
                if pfield in row.keys():
                    value = row[pfield]
                    if isinstance(value, str):
                        found_ontology = re.search(r"(.+) \[\w+:.*", value)
                        if found_ontology:
                            # remove the ontology data from item value
                            value = found_ontology.group(1)
                        s_dict[pfield] = value
            # include the fixed value
            fixed_value = self.config_json.get_topic_data(
                "upload_database", "iskylims_fixed_values"
            )
            for prop, val in fixed_value.items():
                s_dict[prop] = val
            all_iskylims_fields = s_project_fields + s_fields
            sid = row.get("sequencing_sample_id", row.get("sequence_file_R1"))
            for missing in list(set(list(row.keys())) - set(all_iskylims_fields)):
                self.log.debug(f"Field {missing} not found in Iskylims in sample {sid}")
            sample_list.append(s_dict)
        return sample_list

    def get_iskylims_fields_sample(self):
        """2 requests are sent to iSkyLIMS. One for getting the sample fields
        These fields are mapped using the ontology.
        The second request is for getting the sample project fields. These are
        mapped using the label value.
        """

        sample_fields = {}
        s_project_fields = []
        # get the ontology values for mapping values in sample fields
        ontology_dict = self.get_schema_ontology_values()
        sample_url = self.platform_settings["iskylims"]["url_sample_fields"]
        sample_fields_raw = self.platform_rest_api.get_request(sample_url, "", "")

        if "ERROR" in sample_fields_raw:
            logtxt1 = f"Unable to fetch data from {self.platform}."
            logtxt2 = f" Received error {sample_fields_raw['ERROR']}"
            self.logsum.add_error(entry=str(logtxt1 + logtxt2))
            stderr.print(f"[red]{logtxt1 + logtxt2}")
            sys.exit(1)

        for _, values in sample_fields_raw["DATA"].items():
            if "ontology" in values:
                try:
                    property = ontology_dict[values["ontology"]]
                    # sample_fields has a key, the label in metadata, and as value
                    # the field name for the sample
                    sample_fields[property] = values["field_name"]
                except KeyError as e:
                    self.logsum.add_warning(entry=f"Error mapping ontology {e}")
                    # stderr.print(f"[red]Error mapping ontology {e}")
            else:
                # for the ones that do not have ontology label in the sample field
                # and have an empty value: sample_fields[key] = ""
                logtxt = f"No ontology found for {values.get('field_name')}"
                self.logsum.add_warning(entry=(logtxt))
        # fetch label for sample Project
        s_project_url = self.platform_settings["iskylims"]["url_project_fields"]
        param = self.platform_settings["iskylims"]["param_sample_project"]
        p_name = self.platform_settings["iskylims"]["project_name"]
        s_project_fields_raw = self.platform_rest_api.get_request(
            s_project_url, param, p_name
        )
        if "ERROR" in s_project_fields_raw:
            logtxt1 = f"Unable to fetch data from {self.platform}."
            logtxt2 = f" Received error {s_project_fields_raw['ERROR']}"
            self.logsum.add_error(entry=str(logtxt1 + logtxt2))
            return
        else:
            self.log.info("Fetched sample project fields from iSkyLIMS")
            stderr.print("[blue] Fetched sample project fields from iSkyLIMS")
        for field in s_project_fields_raw["DATA"]:
            s_project_fields.append(field["sample_project_field_name"])
        return [sample_fields, s_project_fields]

    def map_relecov_sample_data(self):
        """Select the values from self.json_data"""
        field_values = []
        r_fields = self.config_json.get_topic_data(
            "upload_database", "relecov_sample_metadata"
        )

        for row in self.json_data:
            s_dict = {}
            for r_field in r_fields:
                if r_field in row:
                    s_dict[r_field] = row[r_field]
                else:
                    s_dict[r_field] = None
            field_values.append(s_dict)
        return field_values

    def clean_ambiguous_value(self, value):
        """Replace ambiguous values by default if value is not required."""
        if isinstance(value, str):
            if value.strip() in {"", "NA", "None"}:
                return "Not Provided"
        elif value is None:
            return "Not Provided"
        return value

    def update_database(self, field_values, post_url):
        """Send the request to update database"""
        post_url = self.platform_settings[self.platform][post_url]
        suces_count = 0
        request_count = 0
        for chunk in field_values:
            req_sample = ""
            request_count += 1
            if "sample_name" in chunk:
                stderr.print(
                    f"[blue] sending request for sample {chunk['sample_name']}"
                )
                req_sample = chunk["sample_name"]
            elif "sequencing_sample_id" in chunk:
                stderr.print(
                    f"[blue] sending request for sample {chunk['sequencing_sample_id']}"
                )
                req_sample = chunk["sequencing_sample_id"]
            self.logsum.feed_key(sample=req_sample)
            result = self.platform_rest_api.post_request(
                json.dumps(chunk),
                {"user": self.user, "pass": self.passwd},
                post_url,
            )
            if "ERROR" in result:
                if result["ERROR"] == "Server not available":
                    # retry to connect to server
                    for i in range(10):
                        # wait 5 sec before resending the request
                        time.sleep(5)
                        result = self.platform_rest_api.post_request(
                            json.dumps(chunk),
                            {"user": self.user, "pass": self.passwd},
                            self.platform_settings[post_url],
                        )
                        if "ERROR" not in result:
                            break
                    if i == 9 and "ERROR" in result:
                        logtxt = f"Unable to sent the request to {post_url}"
                        self.logsum.add_error(entry=logtxt, sample=req_sample)
                        stderr.print(f"[red]{logtxt}")
                        continue

                elif "is not defined" in result["ERROR_TEST"].lower():
                    error_txt = result["ERROR_TEST"]
                    logtxt = f"Sample {req_sample} failed in {post_url}: {error_txt}"
                    self.logsum.add_error(entry=logtxt, sample=req_sample)
                    stderr.print(f"[yellow]Warning: {logtxt}")
                    continue
                elif "already defined" in result["ERROR_TEST"].lower():
                    logtxt = f"Request to {post_url} already defined"
                    self.logsum.add_warning(entry=logtxt, sample=req_sample)
                    stderr.print(f"[yellow]{logtxt} for sample {req_sample}")
                    continue
                else:
                    logtxt = f"Error {result['ERROR']} in request to {post_url}"
                    self.logsum.add_error(entry=logtxt, sample=req_sample)
                    stderr.print(f"[red]{logtxt}")
                    continue
            self.log.info(
                "stored data in %s iskylims for sample %s",
                self.platform,
                req_sample,
            )
            stderr.print(f"[green] Successful request for {req_sample}")
            suces_count += 1
        if request_count == suces_count:
            stderr.print(
                f"All {self.type_of_info} data sent sucessfuly to {self.platform}"
            )
        else:
            logtxt = "%s of the %s requests were sent to %s"
            self.logsum.add_warning(
                entry=logtxt % (suces_count, request_count, self.platform),
                sample=req_sample,
            )
            stderr.print(
                f"[yellow]{logtxt % (suces_count, request_count, self.platform)}"
            )
        return

    def store_data(self, type_of_info, server_name):
        """Collect data from json file and split them to store data in iSkyLIMS
        and in Relecov Platform
        """
        map_fields = {}

        if type_of_info == "sample":
            if server_name == "iskylims":
                self.log.info("Getting sample fields from %s", server_name)
                stderr.print(f"[blue] Getting sample fields from {server_name}")
                sample_fields, s_project_fields = self.get_iskylims_fields_sample()
                self.log.info("Selecting sample fields")
                stderr.print("[blue] Selecting sample fields")
                map_fields = self.map_iskylims_sample_fields_values(
                    sample_fields, s_project_fields
                )
            else:
                stderr.print("[blue] Selecting sample fields")
                map_fields = self.map_relecov_sample_data()
            post_url = "store_samples"

        elif type_of_info == "bioinfodata":
            post_url = "bioinfodata"
            map_fields = [
                {k: self.clean_ambiguous_value(v) for k, v in row.items()}
                for row in self.json_data
            ]

        elif type_of_info == "variantdata":
            post_url = "variantdata"
            map_fields = self.json_data

        self.update_database(map_fields, post_url)
        stderr.print(f"[green]Upload process to {self.platform} completed")

    def start_api(self, platform):
        """Open connection torwards database server API"""
        # Get database settings
        try:
            p_settings = self.platform_settings[platform]
        except KeyError as e:
            logtxt = f"Unable to fetch parameters for {platform} {e}"
            stderr.print(f"[red]{logtxt}")
            self.logsum.add_error(entry=logtxt)
            sys.exit(1)
        if self.server_url is None:
            server_url = p_settings["server_url"]
        else:
            server_url = self.server_url
        self.platform = platform
        self.api_url = p_settings["api_url"]
        self.platform_rest_api = RestApi(server_url, self.api_url)
        return

    def update_db(self):
        """Run the update database process with the provided input"""
        if self.full_update is True:
            self.server_name = "iskylims"
            self.type_of_info = "sample"
            self.start_api(self.server_name)
            self.store_data(self.type_of_info, self.server_name)

            self.server_name = "relecov"
            self.start_api(self.server_name)
            for datatype in ["sample", "bioinfodata", "variantdata"]:
                log_text = f"Sending {datatype} data to {self.server_name}"
                self.log.info(log_text)
                stderr.print(log_text)
                self.type_of_info = datatype
                # TODO: Handling for servers with different datatype needs
                if datatype == "variantdata":
                    self.log.info(
                        "Selected %s file for variant data", str(self.long_table_file)
                    )
                    self.json_data = relecov_tools.utils.read_json_file(
                        self.long_table_file
                    )
                self.store_data(datatype, self.server_name)
        else:
            self.start_api(self.platform)
            if self.type_of_info == "variantdata":
                self.log.info(
                    "Selected %s file for variant data", str(self.long_table_file)
                )
                self.json_data = relecov_tools.utils.read_json_file(
                    self.long_table_file
                )
            self.store_data(self.type_of_info, self.platform)
        self.parent_create_error_summary(called_module="update-db")
        return
